#lang sicp

;;;;METACIRCULAR EVALUATOR THAT SEPARATES ANALYSIS FROM EXECUTION
;;;; FROM SECTION 4.1.7 OF STRUCTURE AND INTERPRETATION OF COMPUTER PROGRAMS

;; OP: Modified to run in racket.

(#%require racket/trace)
(#%require "syntax.rkt")
(#%require "environment.rkt")
(#%require "common.rkt")
(#%require "special-forms.rkt")

(#%provide analyzing-driver-loop
           analyzing-eval-program)

;;;SECTION 4.1.7

(define (analyzing-eval exp env)
  ((analyze exp analyze) env))

(define (analyze exp analyze-proc)
  (cond ((self-evaluating? exp) 
         (analyze-self-evaluating exp analyze-proc))
        ((quoted? exp) (analyze-quoted exp analyze-proc))
        ((variable? exp) (analyze-variable exp analyze-proc))
        ((assignment? exp) (analyze-assignment exp analyze-proc))
        ((definition? exp) (analyze-definition exp analyze-proc))
        ((if? exp) (analyze-if exp analyze-proc))
        ((lambda? exp) (analyze-lambda exp analyze-proc))
        ((begin? exp) (analyze-sequence (begin-actions exp) analyze-proc))
        ((cond? exp) (analyze (cond->if exp)) analyze-proc)
        ((let? exp) (analyze (let->combination exp)) analyze-proc)
        ((application? exp) (analyze-application exp analyze-proc))
        (else
         (error "Unknown expression type -- ANALYZE" exp))))


(define input-prompt ";;; MA-Eval input:")
(define output-prompt ";;; MA-Eval value:")

(define (analyzing-driver-loop)
  (define env (setup-environment))
  (define (driver-loop)
    (prompt-for-input input-prompt)
    (let ((input (read)))
      (let ((output (analyzing-eval input env)))
        (announce-output output-prompt)
        (analyzing-user-print output)))
    (driver-loop))
  (driver-loop))

(define (analyzing-user-print object)
  (if (compound-procedure? object)
      (display (list 'compound-procedure
                     (procedure-parameters object)
                     (procedure-body object)
                     '<procedure-env>))
      (display object)))

(define (analyzing-eval-program program)
  (define env (setup-environment))
  (eval-program program analyzing-eval
                analyzing-user-print env))

'ANALYZING-METACIRCULAR-EVALUATOR-LOADED

;; Exercise 4.23: For a sequence containing a single procedure:
;;
;; (In the below, pn is the result of (analyze procn))
;; Text analyser:
;; (analyze-sequence (proc1))
;; -> (loop p1 ())
;; -> p1
;; 
;; Alyssa analyser:
;; (analyze-sequence (proc1))
;; -> (lambda (env) (execute-sequence (p1) env))
;; -> (lambda (env) (p1 env))

;; Here, the second analyser introduces an additional procedure
;; application rather than just returning the result of (analyze
;; proc1). 

;; For two procedures:
;;
;; Text:
;; (analyze-sequence (proc1 proc2))
;; -> (loop p1 (p2))
;; -> (loop (lambda (e1) (p1 e1) (p2 e1)) ())  ;; This procedure
;; application is done during analyze-sequence.
;; -> (lambda (e1) (p1 e1) (p2 e1))
;; 
;; Alyssa analyser:
;; (analyse-sequence (proc1 proc2))
;; -> (lambda (env) (execute-sequence (p1 p2) env))  ;;
;; analyze-sequence returns this procdure (the recursive calls to
;; execute-sequence that follow are not called until the procedure is evaluated).

;; When the procedure is evaluated:
;; -> (lambda (env) (p1 env) (execute-sequence (p2) env))
;; -> (lambda (env) (p1 env) (p2 env))

;; So the result is the same, but the first analyser returns a
;; procedure that returns the result in a single application, whereas
;; the second requires further evaluation to get the result.
