#lang sicp

;; Exercise 4.24: Define a procedure to run a program through an
;; interpreter and time the run.

(#%require "dd-mceval.rkt"
           "analyzing-mceval.rkt")

(#%require (only racket current-process-milliseconds))
(#%require (only racket/base module+))


;; Call proc on args n times and output the average time per run
(define time-proc
  (lambda (iterations proc)
    (let ((start-time (current-process-milliseconds)))
      (let loop ((n iterations))
        (if (> n 0)
            (begin
              (proc)
              (loop (- n 1)))))
      (exact->inexact
       (/ (- (current-process-milliseconds)
             start-time)
          iterations)))))

(define (fact-program x)
  `((define (fact n)
      (if (= n 1) 1
          (* n (fact (- n 1)))))
    (fact ,x)))

(define (fib-program x)
  `((define (fib n)
      (cond ((= n 0) 0)
            ((= n 1) 1)
            (else
             (+ (fib (- n 2))
                (fib (- n 1))))))
    (fib ,x)))

(#%require rackunit)

(module+ test
  (check-equal? (dd-eval-program (fact-program 1)) 1)
  (check-equal? (analyzing-eval-program (fact-program 1)) 1)
  (check-equal? (dd-eval-program (fact-program 5)) 120)
  (check-equal? (analyzing-eval-program (fact-program 5)) 120)
  (check-equal? (dd-eval-program (fact-program 10))
                (analyzing-eval-program (fact-program 10)))

  (check-equal? (dd-eval-program (fib-program 5))
                (analyzing-eval-program (fib-program 5))))

(module+ timing
  (define dd-fib (time-proc 100 (lambda () (dd-eval-program (fib-program 10)))))
  (define analyzing-fib (time-proc 100 (lambda () (analyzing-eval-program (fib-program 10)))))
  (define dd-fact (time-proc 100 (lambda () (dd-eval-program (fact-program 50)))))
  (define analyzing-fact (time-proc 100 (lambda () (analyzing-eval-program (fact-program 50)))))

  (display `("Fib: dd " ,dd-fib " analyzing " ,analyzing-fib " speed-up " ,(/ dd-fib analyzing-fib)))
  (newline)
  (display `("Fact: dd " ,dd-fact " analyzing " ,analyzing-fact " speed-up " ,(/ dd-fib analyzing-fact))))

;; Test results
;; test-eval.rkt> (time-proc 100 eval-program fib-program mce-eval mce-user-print)
;; 203/100
;; test-eval.rkt> (time-proc 100 eval-program fib-program analyzing-eval analyzing-user-print)
;; 18/25

;; test-eval.rkt> (time-proc 100 eval-program fact-program mce-eval mce-user-print)
;; 183/100
;; test-eval.rkt> (time-proc 100 eval-program fact-program analyzing-eval analyzing-user-print)
;; 89/100

;; This demonstrates the substantially better performance of the
;; analyzing evaluator.
